from __future__ import print_function

from __future__ import absolute_import
import ipywidgets as ipw

from subprocess import check_output
from IPython.display import clear_output

from aiida.orm import Code

VALID_AIIDA_CODE_SETUP_ARGUMETNS = {'label', 'selected_computer', 'plugin', 'description',
                                    'exec_path', 'prepend_text', 'append_text'}

def valid_arguments(arguments, valid_arguments):
    result = {}
    for key, value in arguments.items():
        if key in valid_arguments:
            if type(value) is tuple or type(value) is list:
                result[key] = '\n'.join(value)
            else:
                result[key] = value
    return result

def extract_aiidacodesetup_arguments(arguments):
    return valid_arguments(arguments, VALID_AIIDA_CODE_SETUP_ARGUMETNS)


class CodeDropdown(ipw.VBox):
    def __init__(self, input_plugin, text='Select code:', **kwargs):
        """ Dropdown for Codes for one input plugin.

        :param input_plugin: Input plugin of codes to show
        :type input_plugin: str
        :param text: Text to display before dropdown
        :type text: str
        """

        self.input_plugin = input_plugin
        self.codes = {}

        self.dropdown = ipw.Dropdown(description=text, disabled=True)
        self._btn_refresh = ipw.Button(description="Refresh", layout=ipw.Layout(width="70px"))
        self._btn_refresh.on_click(self.refresh)
        self._setup_another = ipw.HTML(value="""<a href=./setup_code.ipynb target="_blank">Setup new code</a>""")
        self.output = ipw.Output()

        children = [ipw.HBox([self.dropdown, self._btn_refresh, self._setup_another]),
                    self.output]

        super(CodeDropdown, self).__init__(children=children, **kwargs)

        self.refresh()

    def _get_codes(self, input_plugin):
        from aiida.orm.querybuilder import QueryBuilder
        from aiida.backends.utils import get_automatic_user
        from aiida.orm import Computer
        current_user = get_automatic_user()

        qb = QueryBuilder()
        qb.append(
            Computer, filters={'enabled': True}, project=['*'], tag='computer')
        qb.append(
            Code,
            filters={
                'attributes.input_plugin': {
                    '==': input_plugin
                },
                'extras.hidden': {
                    "~==": True
                }
            },
            project=['*'],
            has_computer='computer')
        results = qb.all()

        # only codes on computers configured for the current user
        results = [r for r in results if r[0].is_user_configured(current_user)]

        codes = {"{}@{}".format(r[1].label, r[0].name): r[1] for r in results}
        return codes

    def refresh(self, b=None):
        with self.output:
            clear_output()
            self.codes = self._get_codes(self.input_plugin)
            options = list(self.codes.keys())

            self.dropdown.options = options

            if not options:
                print("No codes found for input plugin '{}'.".format(
                    self.input_plugin))
                self.dropdown.disabled = True
            else:
                self.dropdown.disabled = False

    @property
    def selected_code(self):
        try:
            return self.codes[self.dropdown.value]
        except KeyError:
            return None

class AiiDACodeSetup(ipw.VBox):
    """Class that allows to setup AiiDA code"""
    def __init__(self, **kwargs):
        from aiida.common.pluginloader import all_plugins
        from aiidalab_widgets_base.computers import ComputerDropdown

        style = {"description_width":"200px"}

        # list of widgets to be displayed

        self._inp_code_label = ipw.Text(description="AiiDA code label:",
                                       layout=ipw.Layout(width="500px"),
                                       style=style)

        self._computer = ComputerDropdown(layout={'margin': '0px 0px 0px 125px'})

        self._inp_code_description = ipw.Text(placeholder='No description (yet)',
                                              description="Code description:",
                                              layout=ipw.Layout(width="500px"),
                                              style=style)

        self._inp_code_plugin = ipw.Dropdown(options=sorted(all_plugins('calculations')),
                                             description="Code plugin:",
                                             layout=ipw.Layout(width="500px"),
                                             style=style)

        self._exec_path = ipw.Text(placeholder='/path/to/executable',
                                   description="Absolute path to executable:",
                                   layout=ipw.Layout(width="500px"),
                                   style=style)

        self._prepend_text = ipw.Textarea(placeholder='Text to prepend to each command execution',
                                          description='Prepend text:',
                                          layout=ipw.Layout(width="400px"))

        self._append_text = ipw.Textarea(placeholder='Text to append to each command execution',
                                         description='Append text:',
                                         layout=ipw.Layout(width="400px"))

        self._btn_setup_code = ipw.Button(description="Setup code")
        self._btn_setup_code.on_click(self._setup_code)
        self._setup_code_out = ipw.Output()
        children = [ipw.HBox([ipw.VBox([self._inp_code_label,
                                        self._computer,
                                        self._inp_code_plugin,
                                        self._inp_code_description,
                                        self._exec_path]), ipw.VBox([self._prepend_text,
                                                                     self._append_text])]),
                    self._btn_setup_code,
                    self._setup_code_out,
                   ]
        # Check if some settings were already provided
        self._predefine_settings(**kwargs)
        super(AiiDACodeSetup, self).__init__(children, **kwargs)

    def _predefine_settings(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise AttributeError("'{}' object has no attribute '{}'".format(self, key))

    def _setup_code(self, b=None):
        with self._setup_code_out:
            clear_output()
            if self.label is None:
                print("You did not specify code label")
                return
            if not self.exec_path:
                print("You did not specify absolute path to the executable")
                return
            if self.exists():
                print ("Code {}@{} already exists".format(self.label, self.selected_computer.name))
                return
            code = Code(remote_computer_exec=(self.selected_computer, self.exec_path))
            code.label = self.label
            code.description = self.description
            code.set_input_plugin_name(self.plugin)
            code.set_prepend_text(self.prepend_text)
            code.set_append_text(self.append_text)
            code.store()
            code._reveal()
            full_string = "{}@{}".format(self.label, self.selected_computer.name)
            print(check_output(['verdi', 'code', 'show', full_string]))

    def exists(self):
        from aiida.common.exceptions import NotExistent, MultipleObjectsError
        try:
            Code.get_from_string("{}@{}".format(self.label, self.selected_computer.name))
            return True
        except MultipleObjectsError:
            return True
        except NotExistent:
            return False

    @property
    def label(self):
        if len(self._inp_code_label.value.strip()) == 0:
            return None
        else:
            return self._inp_code_label.value

    @label.setter
    def label(self, label):
        self._inp_code_label.value = label

    @property
    def description(self):
        return self._inp_code_description.value

    @description.setter
    def description(self, description):
        self._inp_code_description.value = description

    @property
    def plugin(self):
        return self._inp_code_plugin.value

    @plugin.setter
    def plugin(self, plugin):
        if plugin in self._inp_code_plugin.options:
            self._inp_code_plugin.value = plugin

    @property
    def exec_path(self):
        return self._exec_path.value

    @exec_path.setter
    def exec_path(self, exec_path):
        self._exec_path.value = exec_path

    @property
    def prepend_text(self):
        return self._prepend_text.value

    @prepend_text.setter
    def prepend_text(self, prepend_text):
        self._prepend_text.value = prepend_text

    @property
    def append_text(self):
        return self._append_text.value

    @append_text.setter
    def append_text(self, append_text):
        self._append_text.value = append_text

    @property
    def selected_computer(self):
        return self._computer.selected_computer

    @selected_computer.setter
    def selected_computer(self, selected_computer):
        self._computer.selected_computer = selected_computer



# pylint: disable=unused-import
from aiida import load_dbenv, is_dbenv_loaded
from aiida.backends import settings
if not is_dbenv_loaded():
    load_dbenv(profile=settings.AIIDADB_PROFILE)

from .structures import StructureUploadWidget  # noqa
from .structures_multi import MultiStructureUploadWidget  # noqa
from .codes import CodeDropdown, AiiDACodeSetup, extract_aiidacodesetup_arguments # noqa
from .computers import SshComputerSetup, extract_sshcomputersetup_arguments # noqa
from .computers import AiidaComputerSetup, extract_aiidacomputer_arguments # noqa
from .display import aiidalab_display # noqa

__version__ = "0.3.0b1"



from __future__ import print_function

import os
import ase.io
import ipywidgets as ipw
from ipywidgets import Layout
from IPython.display import clear_output
from fileupload import FileUploadWidget
import tarfile
import zipfile
import tempfile
import nglview


class MultiStructureUploadWidget(ipw.VBox):
    def __init__(self, text="Upload Zip or Tar archive", node_class=None, **kwargs):
        """ Upload multiple structures and store them in AiiDA database.

        :param text: Text to display before upload button
        :type text: str
        :param node_class: AiiDA node class for storing the structure.
            Possible values: 'StructureData', 'CifData' or None (let the user decide).
            Note: If your workflows require a specific node class, better fix it here.
        """

        self.file_upload = FileUploadWidget(text)

        # define the view part of the widget
        self.viewer = nglview.NGLWidget()
        self.selection_slider = ipw.SelectionSlider(
            options=[None,],
            disabled=False,
            orientation='vertical',
            description='Browse structures:',
            readout=False,
            layout = Layout(width='50%'),
        )
        view = ipw.HBox([self.viewer, self.selection_slider])

        # define the action part of the widget
        self.btn_store_all = ipw.Button(description='Store all in AiiDA', disabled=True)
        self.btn_store_selected = ipw.Button(description='Store selected', disabled=True)
        self.structure_description = ipw.Text(placeholder="Description (optional)")
        self.data_format = ipw.RadioButtons(
            options=['StructureData', 'CifData'], description='Data type:')

        # if node_class is predefined, there is no need to select it afterwards
        if node_class is None:
            store = ipw.HBox(
                [self.btn_store_all, self.data_format, self.structure_description, self.btn_store_selected])
        else:
            store = ipw.HBox([self.btn_store_all, self.structure_description, self.btn_store_selected])
            self.data_format.value = node_class

        # define main data objects
        self.structure_ase = None # contains the selected structure in the ase format
        self.structure_nodes = [] # a list that contains all stored structure objects
        self.structure_names = [] # list of uploaded structures

        # put all visual parts in children list and initialize the parent Vbox widget with it
        children = [self.file_upload, view, store]
        super(MultiStructureUploadWidget, self).__init__(
            children=children, **kwargs)

        # attach actions to the buttons
        self.file_upload.observe(self._on_file_upload, names='data')
        self.selection_slider.on_trait_change(self.change_structure, 'value')
        self.btn_store_all.on_click(self._on_click_store_all)
        self.btn_store_selected.on_click(self._on_click_store_selected)

    # function to be called when selection_slider changes
    def change_structure(self):
        if self.selection_slider.value is None:
            pass
        else:
            self.select_structure(filepath=self.selection_slider.value)

    # pylint: disable=unused-argument
    def _on_file_upload(self, change):
        # I redefine both: structure_names and structure_nodes, since we are now uploading a different archive
        self.structure_names = []
        self.structure_nodes = []

        # download an archive and put its content into a file
        archive = tempfile.NamedTemporaryFile(suffix=self.file_upload.filename)
        with open(archive.name, 'w') as f:
            f.write(self.file_upload.data)
        self.archive_name = archive.name

        # create a temporary folder where all the structure will be extracted
        self.tmp_folder = tempfile.mkdtemp()

        # extract tar archive
        if tarfile.is_tarfile(self.archive_name):
            try:
                with tarfile.open(self.archive_name, "r:*", format=tarfile.PAX_FORMAT) as tar:
                    if not tar.getmembers():
                        raise ValueError("The input tar file is empty.")
                    for member in tar.getmembers():
                        tar.extract(path=self.tmp_folder, member=member)
            except tarfile.ReadError:
                raise ValueError("The input tar file is corrupted.")

        # extract zip archive
        elif zipfile.is_zipfile(self.archive_name):
            try:
                with zipfile.ZipFile(self.archive_name, "r", allowZip64=True) as zipf:
                    if not zipf.namelist():
                        raise ValueError("The input zip file is empty.")
                    for member in zipf.namelist():
                        zipf.extract(path=self.tmp_folder, member=member)
            except zipfile.BadZipfile:
                raise ValueError("The input zip file is corrupted.")
        else:
            raise ValueError("The file you provided does not look like Zip or Tar archive")

        # put all extracted files into a list
        for (dirpath, dirnames, filenames) in os.walk(self.tmp_folder):
            for filename in filenames:
                self.structure_names.append(dirpath+'/'+filename)
        if not self.structure_names:
            raise ValueError("Even though the input archive seem not to be empty, it does not contain any file")

        self.structure_names.sort()
        # redefining the options for the slider and its default value
        # together with slider's value update the structure selection also changes,
        # as change_structure() called on slider's value change
        self.selection_slider.options = self.structure_names
        self.selection_slider.value = self.structure_names[0]

    def get_ase(self, filepath):
        file_sub_path = filepath[len(self.tmp_folder)+1:]
        try:
            traj = ase.io.read(filepath, index=":")
        except AttributeError:
            print("Looks like {} file does not contain structure coordinates".format(file_sub_path))
            return None
        if len(traj) > 1:
            print("Warning: Uploaded file {} contained more than one structure. "
                  "I take the first one.".format(file_sub_path))
        return traj[0]

    def get_description(self, structure_ase, filepath):
        formula = structure_ase.get_chemical_formula()
        return "{} ({})".format(formula, filepath.split('/')[-1])

    def select_structure(self, filepath):
        structure_ase = self.get_ase(filepath)
        self.btn_store_all.disabled = False
        self.btn_store_selected.disabled = False

        if structure_ase is None:
            self.structure_ase = None
            self.btn_store_selected.disabled = True
            self.refresh_view()
            return

        self.structure_description.value = self.get_description(structure_ase, filepath)
        self.structure_ase = structure_ase
        self.refresh_view()


    def refresh_view(self):
        viewer = self.viewer
        # Note: viewer.clear() only removes the 1st component
        # pylint: disable=protected-access
        for comp_id in viewer._ngl_component_ids:
            viewer.remove_component(comp_id)
        if self.structure_ase is None:
            return
        viewer.add_component(nglview.ASEStructure(
            self.structure_ase))  # adds ball+stick
        viewer.add_unitcell()

    # pylint: disable=unused-argument
    def _on_click_store_all(self, change):
        self.structure_nodes = []
        # comment this if you are sure that it is safe, and the selection_slider does not interfere with store_structure() function
        self.selection_slider.disabled = True
        for filepath in self.structure_names:
            self.store_structure(filepath)
        # comment this if you are sure that it is safe, and the selection_slider does not interfere with store_structure() function
        self.selection_slider.disabled = False

    # pylint: disable=unused-argument
    def _on_click_store_selected(self, change):
        self.store_structure(self.selection_slider.value, description=self.structure_description.value)

    def store_structure(self, filepath, description=None):
        structure_ase = self.get_ase(filepath)
        if structure_ase is None:
            return

        # determine data source
        if filepath.endswith('.cif'):
            source_format = 'CIF'
        else:
            source_format = 'ASE'

        # perform conversion
        if self.data_format.value == 'CifData':
            if source_format == 'CIF':
                from aiida.orm.data.cif import CifData
                structure_node = CifData(
                    file=filepath,
                    scan_type='flex',
                    parse_policy='lazy')
            else:
                from aiida.orm.data.cif import CifData
                structure_node = CifData()
                structure_node.set_ase(structure_ase)
        else:
            # Target format is StructureData
            from aiida.orm.data.structure import StructureData
            structure_node = StructureData(ase=structure_ase)

            #TODO: Figure out whether this is still necessary for StructureData
            # ensure that tags got correctly translated into kinds
            for t1, k in zip(structure_ase.get_tags(),
                             structure_node.get_site_kindnames()):
                t2 = int(k[-1]) if k[-1].isnumeric() else 0
                assert t1 == t2
        if description is None:
            structure_node.description = self.get_description(structure_ase, filepath)
        else:
            structure_node.description = description
        structure_node.label = ".".join(filepath.split('/')[-1].split('.')[:-1])
        structure_node.store()
        self.structure_nodes.append(structure_node)
        print("Stored in AiiDA: " + repr(structure_node))

    @property
    def node_class(self):
        return self.data_format.value

    @node_class.setter
    def node_class(self, value):
        self.data_format.value = value



import itertools
import numpy as np
import spglib

from ase.spacegroup import crystal
from ase.data import atomic_numbers, atomic_names
from ase.spacegroup import Spacegroup
from numpy.linalg import norm


def ase_to_spgcell(ase_atoms):
    return (ase_atoms.get_cell(),
            ase_atoms.get_scaled_positions(),
            ase_atoms.get_atomic_numbers())
def check_crystal_equivalence(crystal_a, crystal_b):
    """Function that identifies whether two crystals are equivalent"""

    # getting symmetry datasets for both crystals
    cryst_a = spglib.get_symmetry_dataset(ase_to_spgcell(crystal_a), symprec=1e-5, angle_tolerance=-1.0, hall_number=0)
    cryst_b = spglib.get_symmetry_dataset(ase_to_spgcell(crystal_b), symprec=1e-5, angle_tolerance=-1.0, hall_number=0)

    samecell = np.allclose(cryst_a['std_lattice'], cryst_b['std_lattice'], atol=1e-5)
    samenatoms = len(cryst_a['std_positions']) == len(cryst_b['std_positions'])
    samespg = cryst_a['number'] == cryst_b['number']
    
    def test_rotations_translations(cryst_a, cryst_b, repeat):
        cell = cryst_a['std_lattice']
        pristine = crystal('Mg', [(0, 0., 0.)], 
                           spacegroup=int(cryst_a['number']),
                           cellpar=[cell[0]/repeat[0], cell[1]/repeat[1], cell[2]/repeat[2]]).repeat(repeat)

        sym_set_p = spglib.get_symmetry_dataset(ase_to_spgcell(pristine), symprec=1e-5,
                                               angle_tolerance=-1.0, hall_number=0)

        for _,trans in enumerate(zip(sym_set_p['rotations'], sym_set_p['translations'])):
            pnew=(np.matmul(trans[0],cryst_a['std_positions'].T).T + trans[1]) % 1.0
            fulln = np.concatenate([cryst_a['std_types'][:, None], pnew], axis=1)
            fullb = np.concatenate([cryst_b['std_types'][:, None], cryst_b['std_positions']], axis=1)
            sorted_n = np.array(sorted([ list(row) for row in list(fulln) ]))
            sorted_b = np.array(sorted([ list(row) for row in list(fullb) ]))
            if np.allclose(sorted_n, sorted_b, atol=1e-5):
                return True
        return False

    if samecell and samenatoms and samespg:
        cell = cryst_a['std_lattice']
        # we assume there are no crystals with a lattice parameter smaller than 2 A
        rng1 = range(1, int(norm(cell[0])/2.))
        rng2 = range(1, int(norm(cell[1])/2.))
        rng3 = range(1, int(norm(cell[2])/2.))

        for repeat in itertools.product(rng1, rng2, rng3):
            if test_rotations_translations(cryst_a, cryst_b, repeat):
                return True

    return False



from __future__ import print_function
from __future__ import absolute_import

import ase.io
import ipywidgets as ipw
from fileupload import FileUploadWidget
import tempfile
import nglview
from six.moves import zip


class StructureUploadWidget(ipw.VBox):

    DATA_FORMATS = ('StructureData', 'CifData')

    def __init__(self, text="Upload Structure", node_class=None, **kwargs):
        """ Upload a structure and store it in AiiDA database.

        :param text: Text to display before upload button
        :type text: str
        :param node_class: AiiDA node class for storing the structure.
            Possible values: 'StructureData', 'CifData' or None (let the user decide).
            Note: If your workflows require a specific node class, better fix it here.

        """

        self.file_upload = FileUploadWidget(text)
        self.viewer = nglview.NGLWidget()
        self.btn_store = ipw.Button(
            description='Store in AiiDA', disabled=True)
        self.structure_description = ipw.Text(
            placeholder="Description (optional)")

        self.structure_ase = None
        self.structure_node = None
        self.data_format = ipw.RadioButtons(
            options=self.DATA_FORMATS, description='Data type:')

        if node_class is None:
            store = ipw.HBox(
                [self.btn_store, self.data_format, self.structure_description])
        elif node_class not in self.DATA_FORMATS:
            raise ValueError("Unknown data format '{}'. Options: {}".format(
                node_class, self.DATA_FORMATS))
        else:
            self.data_format.value = node_class
            store = ipw.HBox([self.btn_store, self.structure_description])

        children = [self.file_upload, self.viewer, store]

        super(StructureUploadWidget, self).__init__(
            children=children, **kwargs)

        self.file_upload.observe(self._on_file_upload, names='data')
        self.btn_store.on_click(self._on_click_store)

    # pylint: disable=unused-argument
    def _on_file_upload(self, change):
        self.tmp_folder = tempfile.mkdtemp()
        tmp = self.tmp_folder + '/' + self.file_upload.filename
        with open(tmp, 'w') as f:
            f.write(self.file_upload.data)
        self.select_structure(name=self.file_upload.filename)

    def select_structure(self, name):
        structure_ase = self.get_ase(self.tmp_folder + '/' + name)
        self.btn_store.disabled = False
        if structure_ase is None:
            self.structure_ase = None
            self.btn_store.disabled = True
            self.refresh_view()
            return

        self.structure_description.value = self.get_description(
            structure_ase, name)
        self.structure_ase = structure_ase
        self.refresh_view()

    def get_ase(self, fname):
        try:
            traj = ase.io.read(fname, index=":")
        except AttributeError:
            print("Looks like {} file does not contain structure coordinates".
                  format(fname))
            return None
        if len(traj) > 1:
            print(
                "Warning: Uploaded file {} contained more than one structure. I take the first one."
                .format(fname))
        return traj[0]

    def get_description(self, structure_ase, name):
        formula = structure_ase.get_chemical_formula()
        return "{} ({})".format(formula, name)

    def refresh_view(self):
        viewer = self.viewer
        # Note: viewer.clear() only removes the 1st component
        # pylint: disable=protected-access
        for comp_id in viewer._ngl_component_ids:
            viewer.remove_component(comp_id)

        viewer.add_component(nglview.ASEStructure(
            self.structure_ase))  # adds ball+stick
        viewer.add_unitcell()

    # pylint: disable=unused-argument
    def _on_click_store(self, change):
        self.store_structure(
            self.file_upload.filename,
            description=self.structure_description.value)

    def store_structure(self, name, description=None):
        structure_ase = self.get_ase(self.tmp_folder + '/' + name)
        if structure_ase is None:
            return

        # determine data source
        if name.endswith('.cif'):
            source_format = 'CIF'
        else:
            source_format = 'ASE'

        # perform conversion
        if self.data_format.value == 'CifData':
            if source_format == 'CIF':
                from aiida.orm.data.cif import CifData
                structure_node = CifData(
                    file=self.tmp_folder + '/' + name,
                    scan_type='flex',
                    parse_policy='lazy')
            else:
                from aiida.orm.data.cif import CifData
                structure_node = CifData()
                structure_node.set_ase(structure_ase)
        else:  # Target format is StructureData
            from aiida.orm.data.structure import StructureData
            structure_node = StructureData(ase=structure_ase)

            #TODO: Figure out whether this is still necessary for StructureData
            # ensure that tags got correctly translated into kinds
            for t1, k in zip(structure_ase.get_tags(),
                             structure_node.get_site_kindnames()):
                t2 = int(k[-1]) if k[-1].isnumeric() else 0
                assert t1 == t2
        if description is None:
            structure_node.description = self.get_description(
                structure_ase, name)
        else:
            structure_node.description = description
        structure_node.label = ".".join(name.split('.')[:-1])
        structure_node.store()
        self.structure_node = structure_node
        print("Stored in AiiDA: " + repr(structure_node))

    @property
    def node_class(self):
        return self.data_format.value

    @node_class.setter
    def node_class(self, value):
        self.data_format.value = value



from aiida.orm.querybuilder import QueryBuilder
from aiida.orm.data.structure import StructureData

def formula_in_qb(formula):
    # search for existing structures
    qb = QueryBuilder()
    qb.append(StructureData, filters={'extras.formula':formula})
    return [[n[0].get_ase(),n[0].pk] for n in qb.iterall()]



from __future__ import print_function

import importlib
from IPython.display import display

AIIDA_VISUALIZER_MAPPING = {
    'data.parameter.ParameterData.' : 'ParameterDataVisualizer',
    'data.structure.StructureData.' : 'StructureDataVisualizer',
    'data.cif.CifData.'             : 'StructureDataVisualizer',
    'data.folder.FolderData.'       : 'FolderDataVisualizer',
    'data.array.bands.BandsData.'   : 'BandsDataVisualizer',
}

def aiidalab_display(obj, downloadable=True, **kwargs):
    """Display AiiDA data types in Jupyter notebooks.

    :param downloadable: If True, add link/button to download content of displayed AiiDA object.

    Defers to IPython.display.display for any objects it does not recognize.
    """
    from aiidalab_widgets_base import aiida_visualizers
    try:
        visualizer = getattr(aiida_visualizers, AIIDA_VISUALIZER_MAPPING[obj.type])
        display(visualizer(obj, downloadable=downloadable), **kwargs)
    except KeyError:
        display(obj, **kwargs)


from __future__ import print_function
import os

import ipywidgets as ipw

class ParameterDataVisualizer(ipw.HTML):
    """Visualizer class for ParameterData object"""
    def __init__(self, parameter, downloadable=True, **kwargs):
        super(ParameterDataVisualizer, self).__init__(**kwargs)
        import pandas as pd
        # Here we are defining properties of 'df' class (specified while exporting pandas table into html).
        # Since the exported object is nothing more than HTML table, all 'standard' HTML table settings
        # can be applied to it as well.
        # For more information on how to controle the table appearance please visit:
        # https://css-tricks.com/complete-guide-table-element/
        self.value = '''
        <style>
            .df { border: none; }
            .df tbody tr:nth-child(odd) { background-color: #e5e7e9; }
            .df tbody tr:nth-child(odd):hover { background-color:   #f5b7b1; }
            .df tbody tr:nth-child(even):hover { background-color:  #f5b7b1; }
            .df tbody td { min-width: 300px; text-align: center; border: none }
            .df th { text-align: center; border: none;  border-bottom: 1px solid black;}
        </style>
        '''
        pd.set_option('max_colwidth', 40)
        df = pd.DataFrame([(key, value) for key, value
                           in sorted(parameter.get_dict().items())
                          ], columns=['Key', 'Value'])
        self.value += df.to_html(classes='df', index=False) # specify that exported table belongs to 'df' class
                                                            # this is used to setup table's appearance using CSS
        if downloadable:
            import base64
            payload = base64.b64encode(df.to_csv(index=False).encode()).decode()
            fname = '{}.csv'.format(parameter.pk)
            to_add = """Download table in csv format: <a download="{filename}"
            href="data:text/csv;base64,{payload}" target="_blank">{title}</a>"""
            self.value += to_add.format(filename=fname, payload=payload,title=fname)

class StructureDataVisualizer(ipw.VBox):
    """Visualizer class for StructureData object"""
    def __init__(self, structure, downloadable=True, **kwargs):
        import nglview
        self._structure = structure
        viewer = nglview.NGLWidget()
        viewer.add_component(nglview.ASEStructure(self._structure.get_ase()))  # adds ball+stick
        viewer.add_unitcell()
        children = [viewer]
        if downloadable:
            self.file_format = ipw.Dropdown(
                options=['xyz', 'cif'],
                description="File format:",
            )
            self.download_btn = ipw.Button(description="Download")
            self.download_btn.on_click(self.download)
            children.append(ipw.HBox([self.file_format, self.download_btn]))
        super(StructureDataVisualizer, self).__init__(children, **kwargs)

    def download(self, b=None):
        import base64
        from tempfile import TemporaryFile
        from IPython.display import Javascript
        with TemporaryFile() as fobj:
            self._structure.get_ase().write(fobj, format=self.file_format.value)
            fobj.seek(0)
            b64 = base64.b64encode(fobj.read())
            payload = b64.decode()
        js = Javascript(
            """
            var link = document.createElement('a');
            link.href = "data:;base64,{payload}"
            link.download = "{filename}"
            document.body.appendChild(link);
            link.click();
            document.body.removeChild(link);
            """.format(payload=payload,filename=str(self._structure.id)+'.'+self.file_format.value)
        )
        display(js)

class FolderDataVisualizer(ipw.VBox):
    """Visualizer class for FolderData object"""
    def __init__(self, folder, downloadable=True, **kwargs):
        self._folder = folder
        self.files = ipw.Dropdown(
            options=self._folder.get_folder_list(),
            description="Select file:",
        )
        self.text = ipw.Textarea(
            value="",
            description='File content:',
            layout={'width':"900px", 'height':'300px'},
            disabled=False
        )
        self.change_file_view()
        self.files.observe(self.change_file_view, names='value')
        children = [self.files, self.text]
        if downloadable:
            self.download_btn = ipw.Button(description="Download")
            self.download_btn.on_click(self.download)
            children.append(self.download_btn)
        super(FolderDataVisualizer, self).__init__(children, **kwargs)

    def change_file_view(self, b=None):
        with open(self._folder.get_abs_path(self.files.value), "rb") as fobj:
            self.text.value = fobj.read()

    def download(self, b=None):
        import base64
        from IPython.display import Javascript
        with open(self._folder.get_abs_path(self.files.value), "rb") as fobj:
            b64 = base64.b64encode(fobj.read())
            payload = b64.decode()
        js = Javascript(
            """
            var link = document.createElement('a');
            link.href = "data:;base64,{payload}"
            link.download = "{filename}"
            document.body.appendChild(link);
            link.click();
            document.body.removeChild(link);
            """.format(payload=payload,filename=self.files.value)
        )
        display(js)

class BandsDataVisualizer(ipw.VBox):
    """Visualizer class for BandsData object"""
    def __init__(self, bands, **kwargs):
        from bokeh.plotting import figure
        from bokeh.io import show, output_notebook
        output_notebook(hide_banner=True)
        out = ipw.Output()
        with out:
            plot_info = bands._get_bandplot_data(cartesian=True, join_symbol="|")
            y = plot_info['y'].transpose().tolist()
            x = [plot_info['x'] for i in range(len(y))]
            labels = plot_info['labels']
            p = figure(y_axis_label='Dispersion ({})'.format(bands.units))
            p.multi_line(x, y, line_width=2)
            p.xaxis.ticker = [l[0] for l in labels]
            p.xaxis.major_label_overrides = {int(l[0]) if l[0].is_integer() else l[0]:l[1] for l in labels}
            # int(l[0]) if l[0].is_integer() else l[0]
            # This trick was suggested here: https://github.com/bokeh/bokeh/issues/8166#issuecomment-426124290
            show(p)
        children = [out]
        super(BandsDataVisualizer, self).__init__(children, **kwargs)


from __future__ import print_function

import pexpect
import ipywidgets as ipw

from os import path
from copy import copy
from IPython.display import clear_output
from subprocess import check_output, call
from traitlets import Int

from aiida.orm import Computer
from aiida.backends.utils import get_automatic_user, get_backend_type
from aiida.common.exceptions import NotExistent
from aiida.transport.plugins.ssh import parse_sshconfig

if get_backend_type() == 'sqlalchemy':
    from aiida.backends.sqlalchemy.models.authinfo import DbAuthInfo
else:
    from aiida.backends.djsite.db.models import DbAuthInfo

VALID_SSH_COMPUTER_SETUP_ARGUMETNS = {'hostname', 'username', 'proxy_hostname', 'proxy_username'}
VALID_AIIDA_COMPUTER_SETUP_ARGUMETNS = {'name', 'hostname', 'description', 'workdir', 'mpirun_cmd',
                                        'ncpus', 'transport_type', 'scheduler', 'prepend_text', 'append_text'}

def valid_arguments(arguments, valid_arguments):
    result = {}
    for key, value in arguments.items():
        if key in valid_arguments:
            if type(value) is tuple or type(value) is list:
                result[key] = '\n'.join(value)
            else:
                result[key] = value
    return result

def extract_sshcomputersetup_arguments(arguments):
    return valid_arguments(arguments, VALID_SSH_COMPUTER_SETUP_ARGUMETNS)

def extract_aiidacomputer_arguments(arguments):
    return valid_arguments(arguments, VALID_AIIDA_COMPUTER_SETUP_ARGUMETNS)

class SshComputerSetup(ipw.VBox):
    setup_counter = Int(0) # Traitlet to inform other widgets about changes
    def __init__(self, **kwargs):
        style = {"description_width":"200px"}
        computer_image = ipw.HTML('<img width="200px" src="./miscellaneous/images/computer.png">')
        self._inp_username = ipw.Text(description="SSH username:",
                                      layout=ipw.Layout(width="350px"),
                                      style=style)
        self._inp_password = ipw.Password(description="SSH password:",
                                          layout=ipw.Layout(width="130px"),
                                          style=style)
        # Computer ssh settings
        self._inp_computer_hostname = ipw.Text(description="Computer name:",
                                     layout=ipw.Layout(width="350px"),
                                     style=style)
        self._use_proxy = ipw.Checkbox(value=False, description='Use proxy')
        self._use_proxy.observe(self.on_use_proxy_change, names='value')


        # Proxy ssh settings
        self._inp_proxy_address = ipw.Text(description="Proxy server address:",
                                           layout=ipw.Layout(width="350px"),
                                           style=style)
        self._use_diff_proxy_username = ipw.Checkbox(value=False,
                                               description='Use different username and password',
                                               layout={'width': 'initial'})
        self._use_diff_proxy_username.observe(self.on_use_diff_proxy_username_change, names='value')

        self._inp_proxy_username = ipw.Text(value='',
                                            description="Proxy server username:",
                                            layout=ipw.Layout(width="350px"), style=style)
        self._inp_proxy_password = ipw.Password(value='',
                                                description="Proxy server password:",
                                                layout=ipw.Layout(width="138px"),
                                                style=style)
        self._btn_setup_ssh = ipw.Button(description="Setup ssh")
        self._btn_setup_ssh.on_click(self.on_setup_ssh)
        self._setup_ssh_out = ipw.Output()

        # Check if some settings were already provided
        self._predefine_settings(**kwargs)

        # Defining widgets positions
        computer_ssh_box = ipw.VBox([self._inp_computer_hostname,
                                     self._inp_username,
                                     self._inp_password,
                                     self._use_proxy],
                                    layout=ipw.Layout(width="400px"))

        self._proxy_user_password_box = ipw.VBox([self._inp_proxy_username,
                                                  self._inp_proxy_password],
                                                 layout={'visibility':'hidden'})

        self._proxy_ssh_box = ipw.VBox([self._inp_proxy_address,
                                        self._use_diff_proxy_username,
                                        self._proxy_user_password_box],
                                 layout = {'visibility':'hidden','width':'400px'})

        children = [ipw.HBox([computer_image, computer_ssh_box, self._proxy_ssh_box]),
                    self._btn_setup_ssh,
                    self._setup_ssh_out]

        super(SshComputerSetup, self).__init__(children, **kwargs)

    def _predefine_settings(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise AttributeError("'{}' object has no attirubte '{}'".format(self, key))

    def _ssh_keygen(self):
        fn = path.expanduser("~/.ssh/id_rsa")
        if not path.exists(fn):
            print("Creating ssh key pair")
            # returns non-0 if the key pair already exists
            call(["ssh-keygen", "-f", fn, "-t", "rsa", "-N", ""])

    def is_host_known(self, hostname=None):
        if hostname is None:
            hostname = self.hostname
        fn = path.expanduser("~/.ssh/known_hosts")
        if not path.exists(fn):
            return False
        return call(["ssh-keygen", "-F", hostname]) == 0

    def _make_host_known(self, hostname, proxycmd=[]):
        fn = path.expanduser("~/.ssh/known_hosts")
        print("Adding keys from %s to %s"%(hostname, fn))
        hashes = check_output(proxycmd+["ssh-keyscan", "-H", hostname])
        with open(fn, "a") as f:
            f.write(hashes)

    def can_login(self, silent=False):
        if self.username is None: # if I can't find the username - I must fail
            return False
        userhost = self.username+"@"+self.hostname
        if not silent:
            print("Trying ssh "+userhost+"... ", end='')
        # With BatchMode on, no password prompt or other interaction is attempted,
        # so a connect that requires a password will fail.
        ret = call(["ssh", userhost, "-o", "BatchMode=yes", "-o", "ConnectTimeout=5", "true"])
        if not silent:
            print("Ok" if ret==0 else "Failed")
        return ret==0

    def is_in_config(self):
        fn = path.expanduser("~/.ssh/config")
        if not path.exists(fn):
            return False
        cfglines = open(fn).read().split("\n")
        return "Host "+self.hostname in cfglines

    def _write_ssh_config(self, proxycmd=''):
        fn = path.expanduser("~/.ssh/config")
        print("Adding section to "+fn)
        with open(fn, "a") as f:
            f.write("Host "+self.hostname+"\n")
            f.write("User "+self.username+"\n")
            if proxycmd:
                f.write("ProxyCommand ssh -q -Y "+proxycmd+" netcat %h %p\n")
            f.write("ServerAliveInterval 5\n")

    def _send_pubkey(self, hostname, username, password, proxycmd=''):
        from pexpect import TIMEOUT
        timeout = 10
        print("Sending public key to {}... ".format(hostname),end='')
        str_ssh = 'ssh-copy-id {}@{}'.format(username, hostname)
        if proxycmd:
            str_ssh += ' -o "ProxyCommand ssh -q -Y '+proxycmd+' netcat %h %p\n"'
        child = pexpect.spawn(str_ssh)
        try:
            index = child.expect(['s password:', # 0
                                  'ERROR: No identities found', # 1
                                  'All keys were skipped because they already exist on the remote system', # 2
                                  'Could not resolve hostname', # 3
                                  pexpect.EOF],timeout=timeout) # final
        except TIMEOUT:
            print ("Exceeded {} s timeout".format(timeout))
            return False

        if index == 0:
            child.sendline(password)
            try:
                child.expect("Now try logging into",timeout=5)
            except:
                print("Failed")
                print("Please check your username and/or password")
                return False
            print("Ok")
            child.close()
            return True
        elif index == 1:
            print("Failed")
            print("Looks like the key pair is not present in ~/.ssh folder")
            return False
        elif index == 2:
            print("Keys are already there")
            return True
        elif index == 3:
            print("Failed")
            print("Unknown hostname")
            return False
        else:
            print ("Failed")
            print ("Unknown problem")
            print (child.before, child.after)
            child.close()
            return False

    def _configure_proxy(self, password, proxy_password):
        # if proxy IS required
        if self._use_proxy.value:
            # again some standard checks if proxy server parameters are provided
            if self.proxy_hostname is None: # hostname
                print("Please specify the proxy server hostname")
                return False, ''
            # if proxy username and password must be different from the main computer - they should be provided
            if self._use_diff_proxy_username.value:
                # check username
                if not self.proxy_username is None:
                    proxy_username = self.proxy_username
                else:
                    print("Please specify the proxy server username")
                    return False, ''
                # check password
                if len(proxy_password.strip()) == 0:
                    print("Please specify the proxy server password")
                    return False, ''
            else: # if username and password are the same as for the main computer
                proxy_username = self.username
                proxy_password = password
            # make proxy server known
            if not self.is_host_known(self.proxy_hostname):
                self._make_host_known(self.proxy_hostname)
            # Finally trying to connect
            if self._send_pubkey(self.proxy_hostname, proxy_username, proxy_password):
                return True, proxy_username + '@' + self.proxy_hostname
            else:
                print ("Could not send public key to {} (proxy server).".format(self.proxy_hostname))
                return False, ''
        # if proxy is NOT required
        else:
            return True, ''

    def on_setup_ssh(self, b):
        """ATTENTION: modifying the order of operations in this function can lead to unexpected problems"""
        with self._setup_ssh_out:
            clear_output()
            self._ssh_keygen()

            #temporary passwords
            password = self.__password
            proxy_password = self.__proxy_password

            # step 1: if hostname is not provided - do not do anything
            if self.hostname is None: # check hostname
                print("Please specify the computer hostname")
                return

            # step 2: check if password-free access was enabled earlier
            if self.can_login():
                print ("Password-free access is already enabled")
                # it can still happen that password-free access is enabled
                # but host is not present in the config file - fixing this
                if not self.is_in_config():
                    self._write_ssh_config() # we do not use proxy here, because if computer
                    # can be accessed without any info in the config - proxy is not needed.
                    self.setup_counter += 1 # only if config file has changed - increase setup_counter
                return

            # step 3: if can't login already, chek whether all required information is provided
            if self.username is None: # check username
                print("Please enter your ssh username")
                return
            if len(password.strip()) == 0: # check password
                print("Please enter your ssh password")
                return

            # step 4: get the right commands to access the proxy server (if provided)
            success, proxycmd = self._configure_proxy(password, proxy_password)
            if not success:
                return       
            
            # step 5: make host known by ssh on the proxy server
            if not self.is_host_known():
                self._make_host_known(self.hostname,['ssh']+[proxycmd] if proxycmd else [])

            # step 6: sending public key to the main host
            if not self._send_pubkey(self.hostname, self.username, password, proxycmd):
                print ("Could not send public key to {}".format(self.hostname))
                return

            # step 7: modify the ssh config file if necessary
            if not self.is_in_config():
                self._write_ssh_config(proxycmd=proxycmd)
            # TODO: add a check if new config is different from the current one. If so
            # infrom the user about it.

            # step 8: final check
            if self.can_login():
                self.setup_counter += 1
                print("Automatic ssh setup successful :-)")
                return
            else:
                print("Automatic ssh setup failed, sorry :-(")
                return

    def on_use_proxy_change(self, b):
        if self._use_proxy.value:
            self._proxy_ssh_box.layout.visibility = 'visible'
        else:
            self._proxy_ssh_box.layout.visibility = 'hidden'
            self._use_diff_proxy_username.value = False

    def on_use_diff_proxy_username_change(self, b):
        if self._use_diff_proxy_username.value:
            self._proxy_user_password_box.layout.visibility = 'visible'
        else:
            self._proxy_user_password_box.layout.visibility = 'hidden'

# Keep this function only because it might be used later.
# What it does: looks inside .ssh/config file and loads computer setup from 
# there (if present)
#     def _get_from_config(self, b):
#         config = parse_sshconfig(self.hostname)
#         if 'user' in config:
#             self._inp_username.value = config['user']
#         else:
#             self._inp_username.value = ''
#         if 'proxycommand' in config:
#             self._use_proxy.value = True
#             proxy = ''.join([ s for s in config['proxycommand'].split() if '@' in s])
#             username, hostname = proxy.split('@')
#             self._inp_proxy_address.value = hostname
#             if username != self.username:
#                 self._use_diff_proxy_username.value = True
#                 self.proxy_username = username
#         else:
#             self._use_proxy.value = False

    @property
    def __password(self):
        """Returning the password and immediately destroying it"""
        passwd = copy(self._inp_password.value)
        self._inp_password.value = ''
        return passwd

    @property
    def __proxy_password(self):
        """Returning the password and immediately destroying it"""
        passwd = copy(self._inp_proxy_password.value)
        self._inp_proxy_password.value = ''
        return passwd

    @property
    def hostname(self):
        if len(self._inp_computer_hostname.value.strip()) == 0: # check hostname
            return None
        else:
            return self._inp_computer_hostname.value

    @hostname.setter
    def hostname(self, hostname):
        self._inp_computer_hostname.value = hostname

    @property
    def username(self):
        """Loking for username in user's input and config file"""
        if len(self._inp_username.value.strip()) == 0: # if username provided by user
            if not self.hostname is None:
                config = parse_sshconfig(self.hostname)
                if 'user' in config: # if username is present in the config file
                    return config['user']
            else:
                return None
        else:
            return self._inp_username.value

    @username.setter
    def username(self, username):
        self._inp_username.value = username

    @property
    def proxy_hostname(self):
        if len(self._inp_proxy_address.value.strip()) == 0:
            return None
        else:
            return self._inp_proxy_address.value

    @proxy_hostname.setter
    def proxy_hostname(self, proxy_hostname):
        self._use_proxy.value = True
        self._inp_proxy_address.value = proxy_hostname

    @property
    def proxy_username(self):
        if len(self._inp_proxy_username.value.strip()) == 0:
            return None
        else:
            return self._inp_proxy_username.value

    @proxy_username.setter
    def proxy_username(self, proxy_username):
        self._use_proxy.value = True
        self._use_diff_proxy_username.value = True
        self._inp_proxy_username.value = proxy_username

class AiidaComputerSetup(ipw.VBox):
    def __init__(self, **kwargs):
        from aiida.transport import Transport
        from aiida.scheduler import Scheduler
        style = {"description_width":"200px"}

        # list of widgets to be displayed
        self._btn_setup_comp = ipw.Button(description="Setup computer")
        self._btn_setup_comp.on_click(self._on_setup_computer)
        self._inp_computer_name = ipw.Text(value='',
                                           placeholder='Will only be used within AiiDA',
                                           description="AiiDA computer name:",
                                           layout=ipw.Layout(width="500px"),
                                           style=style)
        self._computer_hostname = ipw.Dropdown(description="Select among configured hosts:",
                                               layout=ipw.Layout(width="500px"),
                                               style=style)
        self._inp_computer_description = ipw.Text(value='',
                                                  placeholder='No description (yet)',
                                                  description="Computer description:",
                                                  layout=ipw.Layout(width="500px"),
                                                  style=style)
        self._computer_workdir = ipw.Text(value='/scratch/{username}/aiida_run',
                                          description="Workdir:",
                                          layout=ipw.Layout(width="500px"),
                                          style=style)
        self._computer_mpirun_cmd = ipw.Text(value='mpirun -n {tot_num_mpiprocs}',
                                             description="Mpirun command:",
                                             layout=ipw.Layout(width="500px"),
                                             style=style)
        self._computer_ncpus = ipw.IntText(value=12,
                                           step=1,
                                           description='Number of CPU(s) per node:',
                                           layout=ipw.Layout(width="270px"),
                                           style=style)
        self._transport_type = ipw.Dropdown(value='ssh',
                                            options=Transport.get_valid_transports(),
                                            description="Transport type:",
                                            style=style)
        self._scheduler = ipw.Dropdown(value='slurm',
                                       options=Scheduler.get_valid_schedulers(),
                                       description="Scheduler:",
                                       style=style)
        self._prepend_text = ipw.Textarea(placeholder='Text to prepend to each command execution',
                                          description='Prepend text:',
                                          layout=ipw.Layout(width="400px")
                                         )
        self._append_text = ipw.Textarea(placeholder='Text to append to each command execution',
                                         description='Append text:',
                                         layout=ipw.Layout(width="400px")
                                        )
        self._btn_test = ipw.Button(description="Test computer")
        self._btn_test.on_click(self.test)

        self._setup_comp_out = ipw.Output(layout=ipw.Layout(width="500px"))
        self._test_out = ipw.Output(layout=ipw.Layout(width="500px"))

        # getting the list of available computers
        self.get_available_computers()

        # Check if some settings were already provided
        self._predefine_settings(**kwargs)
        children =[ipw.HBox([ipw.VBox([self._inp_computer_name,
                                       self._computer_hostname,
                                       self._inp_computer_description,
                                       self._computer_workdir,
                                       self._computer_mpirun_cmd,
                                       self._computer_ncpus,
                                       self._transport_type,
                                       self._scheduler]),
                             ipw.VBox([self._prepend_text,
                                       self._append_text])]),
                   ipw.HBox([self._btn_setup_comp, self._btn_test]),
                   ipw.HBox([self._setup_comp_out, self._test_out]),
                  ]
        super(AiidaComputerSetup, self).__init__(children, **kwargs)

    def _predefine_settings(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise AttributeError("'{}' object has no attribute '{}'".format(self, key))

    def get_available_computers(self, b=None):
        fn = path.expanduser("~/.ssh/config")
        if not path.exists(fn):
            return []
        cfglines = open(fn).readlines()
        self._computer_hostname.options = [line.split()[1] for line in cfglines if 'Host' in line]

    def _configure_computer(self):
        """create DbAuthInfo"""
        print("Configuring '{}'".format(self.name))
        sshcfg = parse_sshconfig(self.hostname)
        authparams = {
            'compress': True,
            'gss_auth': False,
            'gss_deleg_creds': False,
            'gss_host': self.hostname,
            'gss_kex': False,
            'key_policy': 'WarningPolicy',
            'load_system_host_keys': True,
            'port': 22,
            'timeout': 60,
        }
        if 'user' in sshcfg:
            authparams['username'] = sshcfg['user']
        else:
            print ("SSH username is not provided, please run `verdi computer configure {}` "
                   "from the command line".format(self.name))
            return
        if 'proxycommand' in sshcfg:
            authparams['proxy_command'] = sshcfg['proxycommand']
        aiidauser = get_automatic_user()
        authinfo = DbAuthInfo(dbcomputer=Computer.get(self.name).dbcomputer, aiidauser=aiidauser)
        authinfo.set_auth_params(authparams)
        authinfo.save()
        print(check_output(['verdi', 'computer', 'show', self.name]))

    def _on_setup_computer(self, b):
        with self._setup_comp_out:
            clear_output()
            if self.name is None: # check hostname
                print("Please specify the computer name (for AiiDA)")
                return
            try:
                computer = Computer.get(self.name)
                print("A computer called {} already exists.".format(self.name))
                return
            except NotExistent:
                pass

            print("Creating new computer with name '{}'".format(self.name))
            computer = Computer(name=self.name)
            computer.set_hostname(self.hostname)
            computer.set_description(self.description)
            computer.set_enabled_state(True)
            computer.set_transport_type(self.transport_type)
            computer.set_scheduler_type(self.scheduler)
            computer.set_workdir(self.workdir)
            computer.set_mpirun_command(self.mpirun_cmd.split())
            computer.set_default_mpiprocs_per_machine(self.ncpus)
            if self._prepend_text.value:
                computer.set_prepend_text(self.prepend_text)
            if self._append_text.value:
                computer.set_append_text(self.append_text)
            computer.store()
            self._configure_computer()

    def test(self, b=None):
        with self._test_out:
            clear_output()
            print(check_output(['verdi', 'computer', 'test', '--traceback', self.name]))

    @property
    def name(self):
        if len(self._inp_computer_name.value.strip()) == 0: # check hostname
            return None
        else:
            return self._inp_computer_name.value

    @name.setter
    def name(self, name):
        self._inp_computer_name.value = name

    @property
    def hostname(self):
        if self._computer_hostname.value is None or len(self._computer_hostname.value.strip()) == 0: # check hostname
            return None
        else:
            return self._computer_hostname.value

    @hostname.setter
    def hostname(self, hostname):
        if hostname in self._computer_hostname.options:
            self._computer_hostname.value = hostname

    @property
    def description(self):
        return self._inp_computer_description.value

    @description.setter
    def description(self, description):
        self._inp_computer_description.value = description

    @property
    def workdir(self):
        return self._computer_workdir.value

    @workdir.setter
    def workdir(self, workdir):
        self._computer_workdir.value = workdir

    @property
    def mpirun_cmd(self):
        return self._computer_mpirun_cmd.value

    @mpirun_cmd.setter
    def mpirun_cmd(self, mpirun_cmd):
        self._computer_mpirun_cmd.value = mpirun_cmd

    @property
    def ncpus(self):
        return self._computer_ncpus.value

    @ncpus.setter
    def ncpus(self, ncpus):
        self._computer_ncpus.value = int(ncpus)

    @property
    def transport_type(self):
        return self._transport_type.value

    @transport_type.setter
    def transport_type(self, transport_type):
        if transport_type in self._transport_type.options:
            self._transport_type.value = transport_type
    @property
    def scheduler(self):
        return self._scheduler.value

    @scheduler.setter
    def scheduler(self, scheduler):
        if scheduler in self._scheduler.options:
            self._scheduler.value = scheduler

    @property
    def prepend_text(self):
        return self._prepend_text.value

    @prepend_text.setter
    def prepend_text(self, prepend_text):
        self._prepend_text.value = prepend_text

    @property
    def append_text(self):
        return self._append_text.value

    @append_text.setter
    def append_text(self, append_text):
        self._append_text.value = append_text

class ComputerDropdown(ipw.VBox):
    def __init__(self,  text='Select computer:', **kwargs):
        """ Dropdown for Codes for one input plugin.

        :param text: Text to display before dropdown
        :type text: str
        """

        self._dropdown = ipw.Dropdown(options=[], description=text, style={'description_width': 'initial'}, disabled=True)
        self._btn_refresh = ipw.Button(description="Refresh", layout=ipw.Layout(width="70px"))

        self._setup_another = ipw.HTML(value="""<a href=./setup_computer.ipynb target="_blank">Setup new computer</a>""",
                                      layout = {'margin': '0px 0px 0px 250px'})
        self._btn_refresh.on_click(self._refresh)
        self.output = ipw.Output()

        children = [ipw.HBox([self._dropdown, self._btn_refresh]),
                    self._setup_another,
                    self.output]

        super(ComputerDropdown, self).__init__(children=children, **kwargs)

        self._refresh()

    def _get_computers(self):
        from aiida.orm.querybuilder import QueryBuilder
        current_user = get_automatic_user()

        qb = QueryBuilder()
        qb.append(
            Computer, filters={'enabled': True}, project=['*'], tag='computer')

        results = qb.all()

        # only computers configured for the current user
        results = [r for r in results if r[0].is_user_configured(current_user)]

        self._dropdown.options = {r[0].name:r[0] for r in results}

    def _refresh(self, b=None):
        with self.output:
            clear_output()
            self._get_computers()
            if not self.computers:
                print("No computers found.")
                self._dropdown.disabled = True
            else:
                self._dropdown.disabled = False

    @property
    def computers(self):
        return self._dropdown.options

    @property
    def selected_computer(self):
        try:
            return self._dropdown.value
        except KeyError:
            return None

    @selected_computer.setter
    def selected_computer(self, selected_computer):
        if selected_computer in self.computers:
            self._dropdown.label = selected_computer




#!/usr/bin/env python

from __future__ import absolute_import
from setuptools import setup, find_packages
import json

# pylint: disable=syntax-error
if __name__ == '__main__':
    # Provide static information in setup.json
    # such that it can be discovered automatically
    with open('metadata.json', 'r') as info:
        metadata = json.load(info)

    with open('setup.json', 'r') as info:
        kwargs = json.load(info)

    setup(
        packages=find_packages(),
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        author=metadata['authors'],
        description=metadata['description'],
        version=metadata['version'],
        **kwargs)
